# GPMaxlik.jl

## Reasonably efficient maximum likelihood estimation for parametric Gaussian process models.

This software package provides respectably optimized methods for evaluating the
negative log-likelihood for mean-zero Gaussian process models with parametric
covariance functions. The primary function is `gnll`, which can also optionally
evaluate the gradient and expected Fisher matrix for given parameter arguments
as well. 

This code is designed to be pretty compartmentalized with regard to separate
computational issues for evaluating the Gaussian log-likelihood and its
derivatives. Primarily, the code accepts arbitrary operator-like objects that
just implement a few specific methods (see `./src/covmatrix.jl` for what they
are). So you can write a struct that wraps some fancy FMM methods for the
solves, for example, and so long as it has the methods needed for the
log-likelihood or derivative you can pass that object right in as the keyword
argument `K`. Because of the stochastic trace estimation described below, you
can avoid all matrix-matrix operations. The primary purpose of the code that is
here is to squeeze out all redundant matrix operations in the evaluation of
log-likelihoods, gradients, and expected Fisher matrices, including in the case
where `k` matrix buffers for derivatives can be provided when you have `k+m`
parameters, so that you need to be careful about the order in which you fill in
elements of the expected Fisher matrix to avoid unnecessarily re-building
derivative matrices.

In order to minimize the matrix-matrix operations required, this software
provides extensive functionality for utilizing the symmetrized stochastic trace
estimators described in [Geoga, Anitescu and Stein
(2019)](https://arxiv.org/abs/1808.03215), which are really very accurate and
effectively exact away from the MLE. The stochastic expected Fisher is probably
accurate to at least single precision even at the MLE. When a stochastic
gradient and expected Fisher matrix are used, all matrix-matrix operations are
eliminated.  This software can also easily be configured to use these stochastic
trace estimators for a fixed number of iterations or until your optimization
routine is reasonably near the MLE in some sense, for example if the norm of the
stochastic gradient is sufficiently small, and then at that time switch to exact
linear algebra.  There is no particular downside that I can see to doing this,
and it effectively just means that you get very near your MLE faster. When you
have many parameters and your initialization point can be incredibly far from
the MLE, this provides some serious speedup. 

As well as providing a tuned `gnll` function, this package also provides a
slightly tuned implementation of trust region-based quasi-Newton optimization.
The code is not particularly sophisticated from an optimization perspective, and
so if you have box constraints, for example, you'll have to handle that in the
way you parameterize the problem. But it is reasonably flexible in the sense
that it's very easy to inject custom functions that run at each iteration, so
that you can use your preferred print output, include custom stopping criteria,
and so on. I also have included convenient bindings for using Ipopt if you
prefer line search methods and are willing to compute exact gradients at least,
but I would encourage everybody to consider the trust region method. Despite its
simplicity, it is pretty competitive.

This software was used to fit a complex 25-parameter model to 17k data points in
[Flexible nonstationary spatio-temporal modeling of high-frequency monitoring
data](https://arxiv.org/abs/2007.11418). I really did not have much luck using
some other popular options, so if your estimation isn't working as well as you'd
like it to, this may be worth trying. And if you want to use stochastic gradient
estimators, you can't use serious optimization software like Ipopt anyway.


**I have written complete examples here for several different types of matrix
structure and optimization methods, so please refer to the files in `./example`
for complete usage demonstrations.**

**If you use this software in your academic work, please cite this software
library (GPMaxlik.jl).**

# Installation

This package is not registered. Please install by opening Julia, pressing the ]
key to enter the Pkg REPL, and entering
````{julia}
Pkg> add https://git.sr.ht/~cgeoga/GPMaxlik.jl
````

