
using GPMaxlik, Ipopt

# Import the ipopt interface function for the Hessian.
#import GPMaxlik: ipopt_efisher!
import GPMaxlik: ipopt_hessian

# Simulate some data to fit. This file also generates the saa vectors, saav.
include("simulate_data.jl")

# Choose inf-norm tolerance for adaptive switching to exact methods (if you want
# that, otherwise leave it set to its default value of zero). I've picked this
# value you'll be able to see the change for this particular data size and
# parameter values and so on. 
GPMaxlik.SETTINGS.GRAD_INFNORM_TOL = 5.0
GPMaxlik.SETTINGS.FISH_INFNORM_TOL = 0.0

# Create a "store" dictionary, which will mitigate Ipopts insistence of having
# separate functions for the gradient and Hessian, even if it almost literally
# always calls the two immediately in sequence. For a stochastic expected Fisher
# matrix, this little trick (which stores extra things on disk, so you don't
# need to fear for your RAM) completely mitigates the extra cost.
include("example_store.jl")
const STORE = SAAStore(Dict{Tuple{UInt64, UInt64}, String}(), true)

# Create the relevant functions for the Ipopt format.
function objective(p, saa_input)
  gnll(pts, simul, kfn, (kd1, kd2), p, saa=saa_input, 
       nll=true, grad=false, fish=false, store=STORE).nll
end

function gradient!(p, g, saa_input)
  nll_call = gnll(pts, simul, kfn, (kd1, kd2), p, saa=saa_input, 
                  nll=false, grad=true, fish=false, store=STORE)
  g .= nll_call.grad
end

function hessian(p, saa_input)
  nll_call = gnll(pts, simul, kfn, (kd1, kd2), p, saa=saa_input, 
                  nll=false, grad=false, fish=true, store=STORE)
  nll_call.fish
end

# Compare exact and stochastic estimates.
for (method, saa) in ((:stochastic, saav), (:exact, nothing))

  # Reset the iteration counter and stochastic derivative flags.
  println()
  GPMaxlik.reset_settings!()

  # set up the problem.
  prob = createProblem(length(pr), [1.0e-5, 1.0e-2], [100.0, 100.0], 0,
                       Float64[], Float64[], 0, div(2*3,2),
                       p->objective(p, saa),
                       (args...)->nothing,
                       (p,g)->gradient!(p, g, saa),
                       (args...)->nothing,
                       (x,m,r,c,o,l,v)->ipopt_hessian(x,m,r,c,o,l,v,
                                                      p->hessian(p, saa),
                                                      Function[], 0))
                      
  addOption(prob, "tol", 1.0e-4)
  addOption(prob, "max_iter", 1000)
  prob.x = [1.0, 5.0] 
  status = solveProblem(prob)
  if status != 0
    println("Optimization for $method estimation failed...")
    continue
  end

  # Get the fisher matrix of the estimate.
  fish = gnll(pts, simul, kfn, (kd1, kd2), prob.x, saa=saa, 
             nll=false, grad=false, fish=true, store=STORE).fish

  # Print output.
  println("Result for $method estimation:")
  for (v, pm) in zip(prob.x, sqrt.(diag(inv(fish))))
    println(round(v, digits=6), " ± ", round(pm*1.96, digits=6))
  end

end

