
using GPMaxlik

# Simulate some data to fit. This file also generates the saa vectors, saav.
include("simulate_data.jl")

# Choose inf-norm tolerance for adaptive switching to exact methods (if you want
# that, otherwise leave it set to its default value of zero). I've picked this
# value you'll be able to see the change for this particular data size and
# parameter values and so on. 
GPMaxlik.SETTINGS.GRAD_INFNORM_TOL = 5.0

# Thin wrappers of the nll function for this more generic trustregion
# implementation, showing off using multiple IID samples:
function prof_nll_only(p)
  gnll(pts, [simul, simul2, simul3], prof_kfn, (prof_kd1,), p, saa=saav,
       nll=true, grad=false, fish=false, vrb=true, profile=true).nll
end

# To use the expected Fisher matrix as a Hessian approximation you could set
# fish=true and change the nothing to nll_call.fish, but the expected Fisher
# matrix for the full likelihood is a much less effective Hessian approximation
# for the profile nll, so this instead opts to use the BFGS Hessian approximation.
function prof_nll_grad_fish(p)
  nll_call = gnll(pts, [simul, simul2, simul3], prof_kfn, (prof_kd1,), p, saa=saav,
                  nll=true, grad=true, fish=false, vrb=true, profile=true)
  (nll_call.nll, nll_call.grad, nothing)
end

# To extract the profile scale parameter:
function prof_scale(p)
  gnll(pts, [simul, simul2, simul3], prof_kfn, (prof_kd1,), p, saa=nothing,
       nll=true, grad=false, fish=false, pscale=true, profile=true).profile_scale
end

# Perform the optimization of the _profile_ likelihood:
output = GPMaxlik.trustregion(prof_nll_only, prof_nll_grad_fish, [5.0], vrb=true)

# Get the MLE and append the profiled scale parameter estimate. Note that the
# scale here looks different than the estimate from the full model, but that is
# because the full model actually finds the square root of the scale so that I
# can use unconstrained optimization.
mle = vcat(prof_scale(output.p), output.p)

if output.status == :FAIL_MAX_ITER
  println("Optimization failed...")
else
  println("Stopping condition: $(output.status)")
  # Print output.
  println(round(mle[1], digits=6), " ± ", nothing)
  for (v, pm) in zip(output.p, sqrt.(diag(inv(output.h))))
    println(round(v, digits=6), " ± ", round(pm*1.96, digits=6))
  end
end

# Add an info note for good measure.
@info "Note that the scale parameter here should be compared with the square root of the point estimates in other examples of fitting the same data because of how the model is parameterized."


