
# This is an example of a "store" object, which will save some internal objects
# in the specific case of computing computing a stochastic gradient but not the
# stochastic Fisher, which Ipopt will do frequently. If you store those internal
# objects---specifically, the derivative-applied SAA vectors---and access them
# when you call the Hessian function, you'll save a pass over the derivative
# matrices, which can be quite a substantial speedup. 
#
# As you can see, this is basically a wrapper over a Dict with a tiny extra flag
# to delete files once they've been used. But if you didn't want that, you could
# also use a raw Dict{...} as in SAAStore.D.
#
# You don't HAVE to use any of this. If you don't want to fiddle, the default
# store value in gnll is Nothing, so this is purely opt-in.

using Serialization

# The input key here is (hash(datav), hash(parms)).
struct SAAStore
  D::Dict{Tuple{UInt64, UInt64}, String}
  delete_after::Bool
end

function Base.delete!(S::SAAStore, key)
  filname = S.D[key]
  rm(filname)
  delete!(S.D, key)
end

function Base.getindex(S::SAAStore, key)
  out = deserialize(S.D[key])
  S.delete_after && delete!(S, key)
  out
end

function Base.setindex!(S::SAAStore, key, value)
  filname_base = mktemp()
  filname = string(deepcopy(filname_base[1]))
  serialize(filname, value)
  S.D[key] = filname
end

function Base.haskey(S::SAAStore, key)
  haskey(S.D, key)
end

