
using GPMaxlik

# Simulate some data to fit. This file also generates the saa vectors, saav.
include("simulate_data.jl")

# Pre-allocate the covariance and derivative matrices. Note that besides these
# three lines, the only thing different between this file and
# example_trustregion.jl is that I pass these objects as kwargs into the nll_*
# functions below. In this case when you look at the timings and alloc
# comparisons you'll see that they're basically identical. But the benefit of
# this interface is that you can provide _ANY_ objects that implement just the
# few methods that GPMaxlik.CovarianceMatrix and GPMaxlik.DerivativeCovarianceMatrix
# do to nll and it will "just work". So with the tiniest bit of wrapper code you
# can drop your H-matrix code or whatever into this reasonably thoughtful nll
# implementation.
const K   = GPMaxlik.build_covmat(pts, kfn, pr)
const Kdv = [GPMaxlik.build_dcovmat(pr, pts) for _ in 1:2]

# Choose inf-norm tolerance for adaptive switching to exact methods (if you want
# that, otherwise leave it set to its default value of zero). I've picked this
# value you'll be able to see the change for this particular data size and
# parameter values and so on. 
GPMaxlik.SETTINGS.GRAD_INFNORM_TOL = 5.0

# Thin wrappers of the nll function for this more generic trustregion
# implementation:
function nll_only(p, saa_input)
  gnll(pts, simul, kfn, (kd1, kd2), p, saa=saa_input,
      nll=true, grad=false, fish=false, vrb=true,
      K=K, Kdv=Kdv).nll
end

# To use the expected Fisher matrix as a Hessian approximation:
function nll_grad_fish(p, saa_input)
  nll_call = gnll(pts, simul, kfn, (kd1, kd2), p, saa=saa_input,
                 nll=true, grad=true, fish=true, vrb=true,
                 K=K, Kdv=Kdv)
  (nll_call.nll, nll_call.grad, nll_call.fish)
end

# To use the BFGS Hessian approximation:
function nll_grad_nofish(p, saa_input)
  nll_call = gnll(pts, simul, kfn, (kd1, kd2), p, saa=saa_input,
                 nll=true, grad=true, fish=false, vrb=true,
                 K=K, Kdv=Kdv)
  (nll_call.nll, nll_call.grad, nothing)
end

# Compare exact and stochastic estimates.
for (method, saa) in ((:stochastic, saav), (:exact, nothing))

  # Reset the iteration counter and stochastic derivative flags.
  println()
  GPMaxlik.reset_settings!()

  # Perform the optimization:
  @time output = GPMaxlik.trustregion(p->nll_only(p, saa), 
                                      p->nll_grad_fish(p, saa),
                                      [1.0, 5.0], vrb=true)

  if output.status == :FAIL_MAX_ITER
    println("Optimization for $method estimation failed...")
    continue
  else
    println("Stopping condition: $(output.status)")
  end

  # Print output.
  println("Result for $method estimation:")
  for (v, pm) in zip(output.p, sqrt.(diag(inv(output.h))))
    println(round(v, digits=6), " ± ", round(pm*1.96, digits=6))
  end

end

