
using GPMaxlik

# Simulate some data to fit. This file also generates the saa vectors, saav.
include("simulate_data.jl")

# Choose inf-norm tolerance for adaptive switching to exact methods (if you want
# that, otherwise leave it set to its default value of zero). I've picked this
# value you'll be able to see the change for this particular data size and
# parameter values and so on. 
GPMaxlik.SETTINGS.GRAD_INFNORM_TOL = 5.0

# Thin wrappers of the nll function for this more generic trustregion
# implementation:
function nll_only(p, saa_input)
  gnll(pts, [simul, simul2, simul3], kfn, (kd1, kd2), p, saa=saa_input,
      nll=true, grad=false, fish=false, vrb=true).nll
end

# To use the expected Fisher matrix as a Hessian approximation:
function nll_grad_fish(p, saa_input)
  nll_call = gnll(pts, [simul, simul2, simul3], kfn, (kd1, kd2), p, saa=saa_input,
                 nll=true, grad=true, fish=true, vrb=true)
  (nll_call.nll, nll_call.grad, nll_call.fish)
end

# Thin wrappers of the nll function for this more generic trustregion
# implementation:
function nll_only_naive(p, saa_input)
  nlls = [gnll(pts, x, kfn, (kd1, kd2), p, saa=saa_input,
              nll=true, grad=false, fish=false, vrb=true).nll
          for x in (simul, simul2, simul3)]
  sum(nlls)
end

# To use the expected Fisher matrix as a Hessian approximation:
function nll_grad_fish_naive(p, saa_input)
  nll_calls = [gnll(pts, x, kfn, (kd1, kd2), p, saa=saa_input,
                   nll=true, grad=true, fish=true, vrb=true)
               for x in (simul, simul2, simul3)]
  (sum(x->x.nll, nll_calls), sum(x->x.grad, nll_calls), sum(x->x.fish, nll_calls))
end

for (_nll_only, _nll_grad_fish, case) in ((nll_only, nll_grad_fish, :smarter),
                                          (nll_only_naive, nll_grad_fish_naive, :naive))

  # Compare exact and stochastic estimates.
  for (method, saa) in ((:stochastic, saav), (:exact, nothing))

    # Reset the iteration counter and stochastic derivative flags.
    println()
    GPMaxlik.reset_settings!()

    # Perform the optimization:
    println("Fitting with $method derivatives and $case computations.")
    @time output = GPMaxlik.trustregion(p->_nll_only(p, saa), 
                                        p->_nll_grad_fish(p, saa),
                                        [1.0, 5.0], vrb=true)

    if output.status == :FAIL_MAX_ITER
      println("Optimization for $method estimation failed...")
      continue
    else
      println("Stopping condition: $(output.status)")
    end

    # Print output.
    println("Result for $method estimation and $case computations:")
    for (v, pm) in zip(output.p, sqrt.(diag(inv(output.h))))
      println(round(v, digits=6), " ± ", round(pm*1.96, digits=6))
    end

  end
end

