
# Set the seed so this script gives the same output each time.
using Random, LinearAlgebra, ForwardDiff, GPMaxlik, StableRNGs

# Choose a covariance function and derivatives. Derivatives here show off a lazy
# option: use AD. This is obviously unnecessary and not performant. But if
# you're prototyping and want as few lines of code as possible, it works.
kfn(x,y,p)  = abs2(p[1])*exp(-norm(x-y)/abs2(p[2]))
kd1(x,y,p)  = ForwardDiff.gradient(pr->kfn(x,y,pr), p)[1] 
kd2(x,y,p)  = ForwardDiff.gradient(pr->kfn(x,y,pr), p)[2] 

# Profiled version of this example function. Note that you could of course not
# write new code and just use vcat to write it as, for example,
#
# prof_kfn(x,y,p) = kfn(x,y,vcat(1.0, p))
# prof_kd1(x,y,p) = kd2(x,y,vcat(1.0, p))
#
# but you would have slower derivatives, and I personally find that code to be
# less clear. Considering how little new code is actually necessary, I prefer
# this more explicit version. And to avoid difficult bugs, I think it is good to
# be disciplined about the number of parameters and number of derivative
# functions be the same. 
prof_kfn(x,y,p) = exp(-norm(x-y)/abs2(p[1]))
prof_kd1(x,y,p) = ForwardDiff.gradient(pr->prof_kfn(x,y,pr), p)[1]

# Simulate some data using that function.
seed = StableRNG(100)
const pr     = [5.0, 2.0]
const pts    = vec(collect.(Iterators.product(1:32, 1:32)))
const covm   = Symmetric([kfn(ptj, ptk, pr) for ptj in pts, ptk in pts])
const covch  = cholesky!(covm)
const simul  = covch.L*randn(seed, length(pts))
const simul2 = covch.L*randn(seed, length(pts))
const simul3 = covch.L*randn(seed, length(pts))

# Fix some SAA vectors to use for each estimation:
const saav  = rand(seed, [-1.0, 1.0], length(pts), 50)

