
# Set the seed so this script gives the same output each time.
using Random, LinearAlgebra, ForwardDiff, GPMaxlik, StableRNGs

# A helper function to build a lower triangular matrix from a vector-shaped
# "packed" form.
function vec7toltri(p)
  out = zeros(eltype(p),3,3)
  out[1,1] = exp(p[2])
  out[2,1] = p[3]
  out[2,2] = exp(p[4])
  out[3,1] = p[5]
  out[3,2] = p[6]
  out[3,3] = exp(p[7])
  LowerTriangular(out)
end

# A more complex covariance function, with seven parameters (for 3D space):
function kfn(x, y, p)
  Lp = vec7toltri(p)
  abs2(p[1])*exp(-norm(Lp\(x-y)))
end

# Get the derivative functions in place:
function coord_deriv(fn, p, j)
  ForwardDiff.derivative(z->fn(vcat(p[1:(j-1)], z, p[(j+1):end])), p[j])
end
const dkfunv = [(x,y,p)->coord_deriv(z->kfn(x,y,z), p, j) for j in 1:7]

# Simulate some data using that function.
seed = StableRNG(100)
const pr     = [2.0, -1.0, 0.5, -0.5, 0.5, -0.5, -1.0]
const mgrid  = range(-1.0, 1.0, length=8)
const pts    = vec(collect.(Iterators.product(mgrid, mgrid, mgrid)))
const _covm  = Symmetric([kfn(ptj, ptk, pr) for ptj in pts, ptk in pts])
const covch  = cholesky!(_covm)
const simul  = covch.L*randn(seed, length(pts))

