
# Glorified re-usable memory buffers. Mostly to serve as a self-documenting
# example for the methods required for more interesting black-box operators to
# interface with the nll function.

# TODO (cg 2020/12/09 13:36): should I export the required methods here?

struct CovarianceMatrix{T}
  M::Cholesky{T, Matrix{T}} 
end

struct DerivativeCovarianceMatrix{T}
  M::Hermitian{T, Matrix{T}} 
end


#
# Methods for covariance matrix:
#

function update!(C::CovarianceMatrix, pts, fun, p) 
  buf = C.M.factors
  Threads.@threads for k in eachindex(pts)
    ptk = pts[k]
    @inbounds for j in 1:k
      buf[j,k] = fun(pts[j], ptk, p)
    end
  end
  cholesky!(Hermitian(buf))
  nothing
end

function sym_factor_solve(C::CovarianceMatrix, x; adj=false)
  adj ? C.M.U\x : adjoint(C.M.U)\x
end

Base.:\(C::CovarianceMatrix, x) = C.M\x

LinearAlgebra.logdet(C::CovarianceMatrix) = logdet(C.M)


#
# Methods for derivative covariance matrix:
#

function update!(Cj::DerivativeCovarianceMatrix, pts, fun, p)
  buf = Cj.M.data
  for k in eachindex(pts)
    @simd for j in 1:k
      @inbounds buf[j,k] = fun(pts[j], pts[k], p)
    end
  end
  nothing
end

Base.:*(Cj::DerivativeCovarianceMatrix, x) = Hermitian(Cj.M)*x


#
# Optional impls:
#

# Note that constructors are optional. If you don't want to implement one in
# this format, then just pass in the objects already constructed. If you do
# that, these methods will never be called.
function build_covmat(pts, fun, pr)
  buf = zeros(eltype(pr), length(pts), length(pts))
  for k in eachindex(pts)
    @simd for j in 1:k
      @inbounds buf[j,k] = fun(pts[j], pts[k], pr)
    end
  end
  CovarianceMatrix(cholesky!(Hermitian(buf)))
end
build_dcovmat(pr, n::Int64) = DerivativeCovarianceMatrix(Hermitian(zeros(eltype(pr), n,n)))
build_dcovmat(pr, pts) = build_dcovmat(pr, length(pts))


function exact_grad_trace(C::CovarianceMatrix, 
                          Cj::DerivativeCovarianceMatrix) 
  K  = C.M
  Kj = Hermitian(Cj.M)
  tr(K\Kj)
end

function exact_efisher_term(C::CovarianceMatrix,
                            Cj::DerivativeCovarianceMatrix,
                            Ck::DerivativeCovarianceMatrix) 
  K  = C.M
  Kj = Hermitian(Cj.M)
  Kk = Hermitian(Ck.M)
  0.5*tr(K\(Kj*(K\Kk)))
end

