  
# A very simple helper function for output depending on what was requested.
function output(l, g, h, ps, nll, grad, fish, pscale)
  out = NamedTuple()
  for (value, name, flag) in zip((l, g, h, ps), 
                                 (:nll, :grad, :fish, :profile_scale), 
                                 (nll, grad, fish, pscale))
    if flag
      if name == :fish
        out = merge(out, (name=>Hermitian(value),))
      else
        out = merge(out, (name=>value,))
      end
    end
  end
  out
end

# A simple helper function to check settings to see if wone should switch from
# stochastic to exact gradient/Fisher objects. If there is a fisher matrix, then
# check things on the Newton step instead of the gradient.
function check_settings!(grad, g, fish, h, saa, vrb, newtonstep=true)
  if (newtonstep && fish && grad)
    try
      gv = h\g 
    catch
      @warn "There was an error from the linear solve for Hessian to obtain Newton step; checking stochastic/exact conditions with gradient instead."
      gv = g
    end
  else
    gv = g
  end
  if grad && !isnothing(saa) && SETTINGS.STOCH_GRAD
    if maximum(abs, gv) < SETTINGS.GRAD_INFNORM_TOL
      vrb && @info "Switching to exact gradients (INFNORM_TOL achieved)."
      SETTINGS.STOCH_GRAD = false
    end
    if SETTINGS.COUNTER > SETTINGS.STOCH_MAXIT
      vrb && @info "Switching to exact gradients (STOCH_MAXIT reached)."
      SETTINGS.STOCH_GRAD = false
    end
  end
  if fish && !isnothing(saa) && SETTINGS.STOCH_FISH
    if maximum(abs, h) < SETTINGS.FISH_INFNORM_TOL
      vrb && @info "Switching to exact Fisher matrices (INFNORM_TOL achieved)."
      SETTINGS.STOCH_FISH = false
    end
    if SETTINGS.COUNTER > SETTINGS.STOCH_MAXIT
      vrb && @info "Switching to exact Fisher matrices (STOCH_MAXIT reached)."
      SETTINGS.STOCH_FISH = false
    end
  end
  nothing
end

function grad_stoch_trace(K_saav, KjK_saav)
  @assert isequal(size(K_saav), size(KjK_saav))
  out = zero(eltype(K_saav))
  for j in 1:size(K_saav, 2)
    v1   = view(K_saav, :, j)
    v2   = view(KjK_saav, :, j)
    out += dot(v1, v2)
  end
  out/size(K_saav, 2)
end

function stoch_efisher_term_diag(Kf, Kj_solved_saav)
  0.5*sum(abs2, Kj_solved_saav)/size(Kj_solved_saav, 2)
end

# To avoid intermediate allocations, this is a manual loop:
function stoch_efisher_term_offdiag(Kf, Kj_solved_saav, Kk_solved_saav)
  @assert isequal(size(Kj_solved_saav), size(Kk_solved_saav))
  out = 0.0
  for k in 1:size(Kj_solved_saav, 2)
    @simd for j in 1:size(Kj_solved_saav, 1)
      @inbounds out += (Kj_solved_saav[j,k] + Kk_solved_saav[j,k])^2
    end
  end
  0.25*out/size(Kj_solved_saav, 2)
end

# Negative log likelihood, with switch for profiled case.
function _nll(Kf, data, dKfd, profile)
  if profile
    return 0.5*logdet(Kf) + 0.5*length(data)*log(dKfd)
  else
    return 0.5*logdet(Kf) + 0.5*dKfd
  end
end

# Solve term in nll gradient, with switch for profile case.
function _grad_solve_term(Kf_solved_data, KjKf_data, dKfd, profile)
  out = dot(Kf_solved_data, KjKf_data)
  if profile
    out *= (length(Kf_solved_data)/dKfd)
  end
  out
end

function grad_entry(Kf, Kf_solved_datav, Kj, profile, 
                    dKfd, solved_saav, Kj_Kfsolved, stoch_grad)
  KjKf_datav = map(x->Kj*x, Kf_solved_datav)
  grad_solve_terms = map(x->_grad_solve_term(x[1], x[2], x[3], profile),
                         zip(Kf_solved_datav, KjKf_datav, dKfd))
  if stoch_grad
    grad_trace_term = grad_stoch_trace(solved_saav, Kj_Kfsolved)
  else
    grad_trace_term = exact_grad_trace(Kf, Kj) # optional impl. 
  end
  0.5*sum(x->grad_trace_term - x, grad_solve_terms)
end

profile_scale(dKfd, len) = dKfd/len

function prepare_operators(K, Kdv, kfn, pts, pr, exact_fish)
  # Build the covariance matrix if not provided, otherwise update it:
  if isnothing(K)
    Kf = build_covmat(pts, kfn, pr)
  else
    Kf = K
    update!(Kf, pts, kfn, pr)
  end
  # Make sure enough deriative matrices are provided. Check a few things in case
  # custom structs are provided.
  default_K_type  = (K isa CovarianceMatrix) || isnothing(K)
  any_deriv_mats  = (length(Kdv) > 0)
  default_Kd_type = any_deriv_mats ? (Kdv[1] isa DerivativeCovarianceMatrix) : default_K_type
  if length(Kdv) < 1 + exact_fish
    if default_Kd_type
      number_needed = max(0, (1+exact_fish)-length(Kdv))
      @info "With these calling parameters, I need to allocate $number_needed more derivative matrices." maxlog=1
      append!(Kdv, map(j->build_dcovmat(pr, pts), 1:number_needed))
    else
      error("You have provided insufficiently many derivative matrices, and because they are your own custom type I can't allocate them. Please pre-allocate at least 2 and try again.")
    end
  end
  (Kf, Kdv)
end

# A stochastic "main loop" to populate things that the gnll call requests. This
# is a pretty ugly function boundary and it could probably use some refactoring.
# I'm still iterating on the process of the cleanest way of handling all the
# branches in those code: stoch or not stoch, fish or not fish, prof or not
# prof, etc.
function stochornofish_loop!(g, h, Kf, Kdv, pts, Kf_solved_datav, dKfd, saa, 
                             solved_saav, dfns, pr, grad, stoch_grad, 
                             profile, fish)
  Kj = Kdv[1] # TODO (cg 2020/12/09 11:31): would using multiple help here?
  # Determine whether to allocate and compute some extra things:
  stochflag = (stoch_grad || fish)
  dapplied_saav = fish ? Vector{Matrix{eltype(pr)}}(undef, length(dfns)) : nothing
  for j in eachindex(dfns)
    update!(Kj, pts, dfns[j], pr)
    # pre-compute Kj*(K\saa) vector once for stochastic case. 
    dapplied_saavj = stochflag ? Kj*solved_saav : nothing
    if grad
      g[j] = grad_entry(Kf, Kf_solved_datav, Kj, profile, dKfd, 
                  solved_saav, dapplied_saavj, stoch_grad)
    end
    # If you need to build the fish, store what you need, additionally applying
    # a symmetric factor solve now that Kj*(K\saa) has been used for the
    # gradient. Because the expected Fisher wants K.L\(Kj*(K\saa)), so we solve
    # with K.L here to avoid doing it repetitively later.
    if fish
      dapplied_saav[j] = sym_factor_solve(Kf, dapplied_saavj)
    end
  end
  # Now that you have everything you need to build the expected Fisher matrix,
  # go ahead and build it if requested.
  if fish
    build_stoch_fish!(h, Kf, dapplied_saav, length(Kf_solved_datav))
  end
  nothing
end

function fish_chunk_update(h, Kf, cJ, Kdv_J, cK, Kdv_K, lendatav)
  for (j, K_J_j) in zip(cJ, Kdv_J)
    for (k, K_K_k) in zip(cK, Kdv_K)
      # For diagonal blocks, make sure you're only filling in one triangle:
      (isequal(cJ, cK) && (k<j)) && continue
      # Fill in the element:
      h[j,k] = exact_efisher_term(Kf, K_J_j, K_K_k)*lendatav
      if j != k
        h[k,j] = h[j,k] # just some defensive programming.
      end
    end
  end
  nothing
end

# A different approach for when an exact fish is requested. This code can
# accomodate a variable number of workspace derivative matrices, and will be
# reasonably efficient with the extra objects if you can spare the space to
# allocate more than the bare minimum of 2.
#
# Currently, there's one more optimization that could be squeezed out of this,
# which is switching the aliases of Kdv_J and Kdv_K at the end of the iteration
# of the outer loop. For example, if each has length 1 and you have two
# parameters, then for the first row you'll compute K_1 and K_2 only once, which
# is good. But then you'll update the buffer again just to fill in the (2,2)
# entry of the matrix, even though you already have K_2 and shouldn't need to
# compute it at all.
function exactfish_loop!(g, h, Kf, Kdv, pts, Kf_solved_datav, dKfd, saa,
                         solved_saav, dfns, pr, grad, stoch_grad, profile)
  # Temporarily, request an even number of Kdv vectors:
  @assert iseven(length(Kdv)) "Please provide an even number of Kdv objects."
  # Since we need to build every entry of this matrix, we split Kdv into two
  # pieces, for J coordinate chunks and K coordinate chunks.
  split_ix = div(length(Kdv)+1, 2)
  Kdv_J    = Kdv[1:split_ix]
  Kdv_K    = Kdv[(split_ix+1):end]
  chunks_J = collect(Iterators.partition(eachindex(pr), length(Kdv_J)))
  chunks_K = collect(Iterators.partition(eachindex(pr), length(Kdv_K)))
  for (cix, cJ) in enumerate(chunks_J[1:(end-1)])
    # Update the J-buffers.
    for (j, K_J_j) in zip(cJ, Kdv_J)
      update!(K_J_j, pts, dfns[j], pr)
    end
    # Compute the gradient if requested:
    if isone(cix) && grad
      for (j, K_J_j) in zip(cJ, Kdv_J)
        if stoch_grad
          dapplied_saavk = K_J_j*solved_saav
        else
          dapplied_saavk = nothing
        end
        g[j] = grad_entry(Kf, Kf_solved_datav, K_J_j, profile, dKfd, 
                          solved_saav, dapplied_saavk, stoch_grad)
      end
    end
    # Update the diagonal block if cix==1:
    if isone(cix)
      fish_chunk_update(h, Kf, cJ, Kdv_J, cJ, Kdv_J, length(Kf_solved_datav))
    end
    # Handle all the off-diagonal blocks now:
    for cK in chunks_K[(cix+1):end]
      for (k, K_K_k) in zip(cK, Kdv_K)
        update!(K_K_k, pts, dfns[k], pr)
      end
      # Sneak in the diagonal block with the Kdv_K bufers:
      fish_chunk_update(h, Kf, cK, Kdv_K, cK, Kdv_K, length(Kf_solved_datav))
      # Sneak in a gradient entry with the Kdv_K bufers::
      if grad
        for (k, K_K_k) in zip(cK, Kdv_K)
          if stoch_grad
            dapplied_saavk = K_K_k*solved_saav
          else
            dapplied_saavk = nothing
          end
          g[k] = grad_entry(Kf, Kf_solved_datav, K_K_k, profile, dKfd, 
                            solved_saav, dapplied_saavk, stoch_grad)
        end
      end
      # Now update the elements of off-diagonal block of h:
      fish_chunk_update(h, Kf, cJ, Kdv_J, cK, Kdv_K, length(Kf_solved_datav))
    end
  end
  nothing
end

# Finish the slightly clever build of the stochastic exected Fisher matrix.
function build_stoch_fish!(h, Kf, dapplied_saav, lendatav)
  # First pass: use dapplied_saav to compute symmetrized estimator.
  for j in 1:size(h,1)   
    dapplj = dapplied_saav[j]
    h[j,j] = stoch_efisher_term_diag(Kf, dapplj) 
    @inbounds for k in 1:(j-1)
      h[j,k] = stoch_efisher_term_offdiag(Kf, dapplj, dapplied_saav[k])
      h[k,j] = h[j,k]
    end
  end
  # Second pass: symmetrization correction (see Geoga et al2019 JCGS)
  for j in 1:size(h,1)   
    for k in 1:size(h,1) 
      if j != k
        h[j,k] -= 0.5*(h[j,j] + h[k,k])
      end
    end
  end
  h .*= lendatav
  nothing
end


"""
Negative log likelihood and its gradient and expected Fisher matrix. Specify
what you want with the kwargs `nll=true`, `grad=true`, and `fish=true`. 

Maybe the only interesting thing here is the keyword arg `saa`. If you provide a
matrix whose columns have the appropriate length, they'll be used as inputs for
a symmetrized stochastic estimator for the gradient and expected Fisher, so long
as they stay above SETTINGS.INFNORM_TOL in infinity norm. After that, the
setting is flipped and future computations are exact. This can really provide a
lot of speedup, and the symmetrized stochastic estimators are quite good away
from the MLE. So when the gradient is large, for example, you really will lose
very little efficiency by using the stochastic derivatives.

Finally, the operations of the large covariance matrices are very
compartmentalized.  You see the optional kwargs of `K`, and `Kdv`, which
optionally allow you to provide a black-box struct that implements just a few
methods for your covariance matrix and its derivative matrices. The file
"./src/covmatrix.jl" is the default implementation of a reasonably careful dense
and exact matrix. But if you wanted to specify a sparse precision matrix or
something more exotic, you could make your own struct, implement the few
required methods, and provide those in the kwargs below.  The `Kd2` object is
only required if you ask for exact Fisher matrices.

Finally, `store` is optionally an object to store some internals of stochastic
derivatives between function calls. It's really only useful for when you're 
interfacing with an optimization software that requires different gradient and
hessian functions, and using a store can completely mitigate the extra cost of
calling those functions separately. See the example files for a full worked 
example and demonstration.

The output is a named tuple of the things that you asked for, with the names
being the same as the keyword arguments.

Signature:

    gnll(pts, data, covfn, derivs_covfn, params, nll, grad, fish,
         saa=nothing, profile=false, vrb=false,
         K=nothing, Kdv=DerivativeCovarianceMatrix[], store=nothing)

See the example files for demonstrations of using this to interface with 
various optimization software suites.
"""
function gnll(pts, datav, kfn, dfns, pr; saa, nll, grad, fish, pscale=false, 
              profile=false, vrb=false, K=nothing, Kdv=DerivativeCovarianceMatrix[],
              store=nothing)
  # Check if there is just one sample provided as a simple vector. If so, modify
  # the call accordingly:
  if (datav isa Vector{T} where{T<:Number})
    return gnll(pts, [datav], kfn, dfns, pr; saa=saa, nll=nll, grad=grad, 
                fish=fish, pscale=pscale, profile=profile, vrb=vrb, 
                K=K, Kdv=Kdv, store=store)
  end
  if !(datav isa Vector{Vector{T}} where{T<:Number})
    @warn "Please provide data as a Vector{Vector}. With current inputs, this might not do what you expect it to do."
  end
  # If nothing is requested, return nothing.
  mapreduce(!, *, (nll, grad, fish, pscale)) && return NamedTuple()
  SETTINGS.COUNTER += 1
  # Check for exact grad and exact fish:
  stoch_grad = SETTINGS.STOCH_GRAD && !isnothing(saa)
  stoch_fish = SETTINGS.STOCH_FISH && !isnothing(saa)
  # Update Kf and make sure there are enough derivative matrices.
  (Kf, Kdv) = prepare_operators(K, Kdv, kfn, pts, pr, (fish && !stoch_fish))
  # Go over several quantities that are useful in several places and/or were
  # requested by the user.
  Kf_solved_datav = map(x->Kf\x,  datav)
  dKfd = map(x->dot(x[1], x[2]), zip(datav, Kf_solved_datav))
  if nll
    el = sum(x->_nll(Kf, x[1], x[2], profile), zip(datav, dKfd))
  else
    el = nothing
  end
  # Check to see if there is a store variable, and if this particular function
  # call means that it should be interacted with.
  hasstore = !isnothing(store)
  store_add_args  = !fish && grad && stoch_grad
  store_take_args = !grad && fish && stoch_fish
  # If this function call is a store_take_args call, then load in that h and
  # skip the rest of the call:
  if hasstore && store_take_args
    key = (hash(datav), hash(pr))
    if haskey(store, key)
      psc = pscale ? dKfd./length(datav[1]) : nothing
      g   = nothing
      h   = Hermitian(store[key])
      check_settings!(grad, g, fish, h, saa, vrb)
      return output(el, g, h, psc, nll, grad, fish, pscale)
    end
  end
  # Otherwise, if it is a store_add_args or otherwise, continue.
  _fish = (hasstore && store_add_args) ? true : fish
  # Allocate for derivatives as necessary:
  psc  = pscale ? dKfd./length(datav[1]) : nothing
  g    = grad   ? Vector{eltype(pr)}(undef, length(dfns)) : nothing
  h    = _fish  ? Matrix{eltype(pr)}(undef, length(dfns), length(dfns)) : nothing
  # If the gradient OR fisher matrix are requested:
  if grad || _fish
    # if stochastic derivative, solve the SAA vectors:
    if stoch_grad || stoch_fish
      solved_saav = sym_factor_solve(Kf, saa, adj=true)
    else
      solved_saav = nothing 
    end
    # In the very specific instance:
    if fish && !stoch_fish
     exactfish_loop!(g, h, Kf, Kdv, pts, Kf_solved_datav, dKfd, saa, solved_saav, 
                     dfns, pr, grad, stoch_grad, profile)
    else
     stochornofish_loop!(g, h, Kf, Kdv, pts, Kf_solved_datav, dKfd, saa, solved_saav,
                         dfns, pr, grad, stoch_grad, profile, _fish)
    end
  end
  # If we need to add to the store, do that:
  if hasstore && store_add_args
    key = (hash(datav), hash(pr))
    setindex!(store, key, Hermitian(h))
  end
  if !fish
    h = nothing
  end
  # Check to see if conditions in the global state for switching from stochastic
  # to deterministic have occured:
  check_settings!(grad, g, fish, h, saa, vrb)
  output(el, g, h, psc, nll, grad, fish, pscale)
end

