
# Thank you to Paul Beckman at Argonne National Lab for improving on the very
# basic implementation of this in KernelMatrices.jl. Paul's improved code was
# used as the template for this implementation here.

subproblem(xv, fx, gx, hx, pv) = fx + dot(gx, pv) + 0.5*dot(pv,hx*pv)

function rho(objfxp, x, fx, gx, hx, p)
  denm = subproblem(x,fx,gx,hx,zeros(length(p))) - subproblem(x,fx,gx,hx,p)
  (fx-objfxp)/denm
end

function solve_subproblem_exact(g, B::Hermitian, del)
  lmin = max(0.0, -real(eigmin(B)))
  lval = lmin + 0.1
  pl   = zeros(length(g))
  for _ in 0:10
    Bi    = cholesky(B + I*lval)
    pl    = -(Bi\g)
    lval += abs2(norm(pl)/norm(Bi.U\pl))*(norm(pl)-del)/del
    (lval < lmin + 1.0e-10) && break
  end
  return pl
end

"""
A simple trust region method with a few twists to try and be maximally flexible.
The required functions are \\
  (1) objective \\
  (2) objective+gradient [+hessian or third return value of nothing for BFGS] \\
  (3) initial input vector \\
Along with standard keyword args for a trust region method, you can provide the 
arg `iter_funs`, which is a collection of functions that get run for each iteration.
They have a required signature; see `GPMaxlik.convg_info` for an example. But with
this you can add your own stopping conditions, different print output to suit your
needs, and so on. You can provide arbitrary keyword args to _those_ functions via
the `iter_funs_kwargs` argument, which takes an iterable list of pairs.
"""
function trustregion(f, fgh, init; vrb=false, dmax=1.0, dini=0.5, eta=0.125, 
                     rtol=1.0e-8, atol=1.0e-5, gtol=1.0e-4, maxit=200, dcut=1.0e-4, 
                     iter_funs=(convg_info, check_obj_nan), iter_funs_kwargs=tuple())
  # initialize:
  xv                = deepcopy(init)
  (dl, r1, fxp, fx) = (dini, 0.0, 0, 0.0)
  (ro, gx, lgx)     = (zeros(length(init)) for _ in 1:3)
  hx                = Matrix{Float64}(I, length(init), length(init))
  # main loop:
  for counter in 1:maxit
    vrb && println("($counter) x: $(round.(xv, digits=4))")
    # update the objective, gradient, and Hessian, _IF_ the step is accepted, or
    # if this is the first iteration.
    (fx, _gx, _hx) = try
      if (r1 > eta) || isone(counter)
        fgh(xv) # _hx can be nothing here.
      else
        (fx, gx, hx)
      end
    catch er
      @warn "Error in calling fgh function. Breaking argument:"
      display(xv)
      rethrow()
    end
    # gradient update:
    gx .= _gx
    # Hessian update:
    if isnothing(_hx) # Use BFGS
      vrb && @info "No Hessian provided, using BFGS approximation." maxlog=1
      if (r1 > eta)
        # Perform quasi-Newton update if the step was accepted
        yv   = gx .- lgx
        zv   = hx*ro
        hx .+= yv*yv'/dot(yv, ro) .- zv*zv'/dot(zv, ro)
      end
      lgx .= gx # Set last gradient to current gradient
    else
      hx .= _hx
    end
    # Loop over optional functions to be called each iteration:
    for fn in iter_funs
      (err_flag, ret) = fn(counter, fx, xv, gx, hx, dl; iter_funs_kwargs...)
      err_flag && return ret
    end
    println()
    # Solve the corresponding sub-problem:
    ro = try
      solve_subproblem_exact(gx, Hermitian(hx), dl)
    catch
      @warn "Error in solving the subproblem. Returning potentially informative objects."
      return (fx=fx, gx=gx, hx=hx, x=xv, dl=dl, r1=r1)
    end
    fxp = f(xv.+ro) 
    r1  = rho(fxp, xv, fx, gx, hx, ro)
    # Compute the Newton step:
    nsx = try
      hx\gx
    catch
      @warn "Error in computing the Newton step. Returning potentially informative objects. Check your Hessian."
      return (fx=fx, g=gx, h=hx, x=xv, dl=dl, r1=r1)
    end
    # Perform tests on solution of sub-problem:
    (r1 < 0.25) && (dl = 0.25*norm(ro))
    (r1 > 0.75) && (dl = min(2.0*norm(ro), dmax))
    (r1 > eta)  && (xv .+= ro)
    # Perform tests for relative and absolute tolerances being reached:
    convg = isconvg(fx, fxp, nsx, gtol, rtol, atol, dl, dcut)
    convg[1] && return (f=fxp, p=xv, g=gx, h=hx, c=counter, status=convg[2])
  end
  return (f=fxp, p=xv, g=gx, h=hx, c=maxit, status=:FAIL_MAX_ITER)
end

# Check for various convergence criteria:
function isconvg(fx, f0, nsx, gtol, rtol, atol, dl, dcut)
  if gtol > 0.0 && isapprox(norm(nsx, Inf), 0.0, atol=gtol) 
    return (true, :NEWTONSTEP_TOL_REACHED)
  elseif rtol > 0.0 && isapprox(fx, f0, rtol=rtol) 
    return (true, :OBJECTIVE_RTOL_REACHED)
  elseif atol > 0.0 && isapprox(fx, f0, atol=atol)
    return (true, :OBJECTIVE_ATOL_REACHED)
  elseif dl < dcut
    return (true, :REGION_TOO_SMALL)
  else
    return (false, :NOT_CONVERGED)
  end
end

# Useful convergence information, thanks to Paul Beckman.
function convg_info(counter, fx, xv, gx, hx, dl; kwargs...)
  nsx = hx\gx
  println("Objective value:      $fx")
  println("Newton step inf norm: $(norm(nsx, Inf))")
  println("Trust region radius:  $dl")
  (false, nothing)
end

function check_obj_nan(counter, fx, xv, gx, hx, dl; kwargs...)
  if isnan(fx) 
    return (true, (f=fx, p=xv, g=gx, h=hx, c=counter, status=:FAIL_OBJ_NAN))
  end
  (false, nothing)
end

# For internal debugging. Don't mind me.
function deriv_printer(counter, fx, xv, gx, hx, dl; kwargs...)
  println("\n\nGradient:")
  display(gx)
  println("\n\nHessian:")
  display(hx)
  println("\n\n")
  (false, nothing)
end

# Fail out if the condition number of the Hessian is sufficiently large:
function check_hessian_cond(counter, fx, xv, gx, hx, dl; maxcond)
  isinf(maxcond) && return (false, nothing)
  if cond(Hermitian(hx)) > maxcond
    return (true, (f=fx, p=xv, g=gx, h=hx, c=counter, status=:FAIL_HESSIAN_COND))
  else
    return (false, nothing)
  end
end

